/* This is a hexdump program.

   A notable feature is that this program doesn't terminate after
   printing all the dumps to terminal, but instead awaits for a
   reeturn key. This can be useful when this program is invoked in
   an environment such as drag-and-dropping a file to read into
   this program under windows. The program will not terminate and
   the command prompt will still be visible. Sure it is possible
   to set this up with some other hexdump and a batch script, but
   whatever.
*/



#include <stdio.h>

/* The version string. */

const char version[] = "0.1.0";

/* The help message. */

const char help_msg[] =	"anon's hexdump version %s\n"
			"usage:\n"
			"\t$ %s <option> <filename>\n"
			"\twhere <filename> is the name of the file you want to view and\n"
			"\t<option> can be one of the following:\n"
			"\t\t-h\tprint this help message\n"
			"\t\t-b\toutput in binary mode\n";

/* The error messages. */

const char err_no_input[] = "%s: error: no input\n";
const char err_no_file_given[] = "%s: error: no file name given\n";
const char err_cannot_open_file[] = "%s: error: failed to open file %s\n";

/* An empty pointer as a global variable to store the name of the program so that
   this name is available to functions other than main. */
char * program_name;

void dumpPrintableChars(char buffer[], unsigned int buf_len){

	/* Function for printing the printable characters of a hexdump. Actually, some
	   other printable characters such as newline are excluded too, because they can
	   interfere with the formatting. This function works best for hexdumps (due to the
	   logic of computing missing collumn space), but may be adjusted to work with
	   bindumps too.*/

	int cnt = 0;	// Kind of a general purpose counter variable used in loops.

	if(buf_len < 16){
		/* The buffer has less than 16 characters. We must add some spaces to
		   the output so that the characters are in the same column as the previously
		   printed characters. */


		/* This is the number of spaces to add. Logically, it should be different for
		   if (1 <= buf_len <= 8) and for if (9 <= buf_len <= 16), however, the space
		   in between the 8 byte segments gets printed regardless of the value of buf_len,
		   so things kinda cancel out and the below formula works in both cases. */
		cnt = 48 - (3*buf_len);

		while(cnt--){
			putchar(' ');
		}

		cnt = 0;	// Return cnt back to the state it was in prior to us using it above.
	}

	printf(" |");

	/* Print the characters. */

	while(cnt < buf_len){
		if(buffer[cnt] >= 0x20 && buffer[cnt] <= 0x7E){	// Is a printable character: print the character.
			putchar(buffer[cnt]);
		}
		else{	// Is not a printable character: print a dot.
			putchar('.');
		}
		cnt++;
	}

	printf("|\n");
}

#define HEXDUMP_SEG_SIZE 16

int dumpFileHex(char filename[]){

	/* Function responsible for producing a hexdump.
	   Returns 0 if everything ok and 1 if error. */

	unsigned int bytes_read;

	unsigned int address = 0;

	/* XXX If this is not unsigned, printf will print 32 bit hex values for negative bytes,
	   which I think it shouldn't do as printing in hex assumes it is unsigned... */
	unsigned char buffer[HEXDUMP_SEG_SIZE] = {};

	FILE * fp = fopen(filename, "rb");	// Open the file.

	if(fp){	// Check if the file was opened successfully.

		/* Read 16-byte segments from the input file and display them
		   as a hexdump while EOF is not reached. */

		do{

			bytes_read = fread(buffer, 1, HEXDUMP_SEG_SIZE, fp);	// Read a 16-byte segment from the file.
			if(bytes_read == 0){
				// If EOF is reached, just break out of the loop.
				break;
			}

			/* Just use printf for a hexdump. */

			printf("%08X  ", address);	// Print the address.
			int cnt = 0;
			while(cnt < 8 && cnt < bytes_read){	// For the first part of the line.
				printf("%02X ", buffer[cnt++]);
			}
			printf(" ");	// For the spaces in the middle of the line.
			while(cnt < bytes_read){	// For the second part of the line.
				printf("%02X ", buffer[cnt++]);
			}
			dumpPrintableChars(buffer, bytes_read);	// Display the printable characters.

			address += bytes_read;

		} while(bytes_read == HEXDUMP_SEG_SIZE);

		// Display the size of the file.
		printf("file size: %d byte(s)\n", address);

		fclose(fp);
	}
	else{
		fprintf(stderr, err_cannot_open_file, program_name, filename);
		return 1;
	}
	return 0;
}

#define BINDUMP_SEG_SIZE 6

int dumpFileBin(char filename[]){

	/* Function responsible for producing a bindump.
   	   Returns 0 if everything ok and 1 if error. */

	unsigned int bytes_read;

	unsigned int address = 0;

	unsigned char buffer[BINDUMP_SEG_SIZE] = {};

	FILE * fp = fopen(filename, "rb");	// Open the file.

	if(fp){	// Check if the file was opened successfully.

		/* Read segments from the input file and display them
		   as a bindump while EOF is not reached. */

		do{

			bytes_read = fread(buffer, 1, BINDUMP_SEG_SIZE, fp);	// Read a segment from the file.
			if(bytes_read == 0){
				// If EOF is reached, just break out of the loop.
				break;
			}

			/* Just use printf for a bindump. */

			printf("%08X  ", address);	// Print the address.
			int cnt = 0;
			while(cnt < bytes_read){
				printf("%08b ", buffer[cnt++]);
			}
			putchar('\n');

			address += bytes_read;

		} while(bytes_read == BINDUMP_SEG_SIZE);

		// Display the size of the file.
		printf("file size: %d byte(s)\n", address);

		fclose(fp);
	}
	else{
		fprintf(stderr, err_cannot_open_file, program_name, filename);
		return 1;
	}
	return 0;
}

/* The main function. */

int main(int argc, char * argv[]){

	program_name = argv[0];

	if(argc > 1){	// Check if there are any input parameters.
		if(argv[1][0] == '-'){	// Check if the first parameter is a flag.

			/* The first parameter is a flag. Process the flag. */

			switch(argv[1][1]){	// argv[1][1] is the first character after the '-'.
				case 'h':
					// Print a help message.
					printf(help_msg, version, program_name);
				break;
				case 'b':
					/* Perform a bindump. */

					if(argc > 2){	// Check if the second parameter is present (should be the file name).
						// Produce a bindump of the file. This function returns 1 if failed.
						if(dumpFileBin(argv[2])){
							return 0;	// We exit the program without waiting for a return key press.
						}

						printf("\nPress enter to terminate program ... ");
						getchar();	// getchar will wait until return is pressed.
					}
					else{
						fprintf(stderr, err_no_file_given, program_name);
					}
				break;
				default:
					printf("Sorry, not implemented yet. Try -h for help.\n");
			}
		}
		else{

			/* The first parameter is not a flag. Treat it as an input file. */

			// Produce a hexdump of the file. This function returns 1 if failed.
			if(dumpFileHex(argv[1])){
				return 0;	// We exit the program without waiting for a return key press.
			}

			printf("\nPress enter to terminate program ... ");
			getchar();	// getchar will wait until return is pressed.
		}
	}
	else{
		fprintf(stderr, err_no_input, program_name);
	}

	return 0;
}
